@@ -24,14 +24,15 @@ try:
   from http.client import HTTPConnection, HTTPSConnection
 except ImportError:
   from httplib import HTTPConnection, HTTPSConnection
-from itertools import groupby
+from itertools import groupby, product
 import json
-from multiprocessing import Process, Queue, Value
+from multiprocessing import Manager, Process, Queue, Value
 from os import path
 try:
   from queue import Empty, Full
 except ImportError:
   from Queue import Empty, Full
+import re
 import socket
 import traceback
 try:
@@ -60,6 +61,7 @@ class Checker():
 
   def __init__(self):
 
+    self.data = {}
     self.links = {}
     self.json_filename = None
 
@@ -113,33 +115,43 @@ class Checker():
     pass
 
   def load_json(self, src):
-    """Load links from a JSON file
+    """Load data and links from a JSON file
 
     src can be a filename or file-like object.
 
     returns src only if it's file and load successfully.
     """
+    filename = None
     if hasattr(src, 'read'):
-      self.links = json.load(src)
+      jsondata = json.load(src)
       src.close()
-      return
+    else:
+      with open(src, 'r') as f:
+        jsondata = json.load(f)
+        filename = src
 
-    with open(src, 'r') as f:
-      self.links = json.load(f)
-      return src
+    # version <= 0.1.1
+    if 'data' not in jsondata or 'links' not in jsondata:
+      self.links = jsondata
+    else:
+      self.data = jsondata['data']
+      self.links = jsondata['links']
+
+    return filename
 
   def save_json(self, dest):
-    """Save links to a JSON file
+    """Save data and links to a JSON file
 
     dest can be a filename or a file-like object, will not be closed by
     save_json if it's a file-like object.
     """
+    jsondata = {'data': self.data, 'links': self.links}
     if hasattr(dest, 'write'):
-      json.dump(self.links, dest)
+      json.dump(jsondata, dest)
       return
 
     with open(dest, 'w') as f:
-      json.dump(self.links, f)
+      json.dump(jsondata, f)
 
   def add_link(self, url, data=None):
     """Add a link
@@ -199,14 +211,40 @@ class Checker():
 
   # =====
 
-  def check_url(self, url, frags=None):
+  RE_ID_NAME = re.compile(r'\b(?:id|name)=(?P<q>[\'"])(.*?)(?P=q)', re.I)
+
+  def _check_url_frag(self, data):
+    """search for fragment in HTML attributes id or name look-likes
+
+    return '200' if found else '###'
+
+    >>> c = Checker()
+    >>> f = c._check_url_frag
+    >>> f(('', 'blah'))
+    '200'
+    >>> f(('foo', 'id="foo"'))
+    '200'
+    >>> f(('foo', 'id="nofoo"'))
+    '###'
+    >>> f(('bar', 'name="bar"'))
+    '200'
+    """
+    frag, body = data
+    if frag == '':
+      return '200'
+    for m in self.RE_ID_NAME.finditer(body):
+      if frag == m.group(2):
+        return '200'
+    return '###'
+
+  def check_url(self, url, frags=None, local_html=None):
     """Check a url
 
-    Returns (final status code(s) in list of string, final redirect url)
+    Returns (final status code(s) in tuple of string, final redirect url)
 
     >>> c = Checker()
     >>> c.check_url('http://example.com') # doctest: +SKIP
-    ('200', 'http://www.iana.org/domains/example')
+    (('200',), 'http://www.iana.org/domains/example')
     """
     MAX_REDIRS = 10
     redirs = 0
@@ -224,6 +262,10 @@ class Checker():
       elif url_comp.scheme == 'https':
         conn = HTTPSConnection(url_comp.netloc)
       else:
+        if not url and frags and local_html:
+          pairs = product(frags, (local_html,))
+          statuses = tuple(map(self._check_url_frag, pairs))
+          break
         status = 'SCH'
         if url_comp.scheme in ('about', 'javascript'):
           status = 'SKP'
@@ -237,16 +279,16 @@ class Checker():
           method = 'GET'
         conn.request(method, p, headers=self.HEADERS)
         resp = conn.getresponse()
-        if resp.status == 200 and frags:
-          rbody = resp.read().decode('utf8')
-          for frag in frags:
-            if 'id="%s"' % frag in rbody:
-              status = '200'
-            elif "id='%s'" % frag in rbody:
-              status = '200'
-            else:
-              status = '###'
-            statuses.append(status)
+        if resp.status == 200 and frags != ('',):
+          # any non text/html result ### as such fragment isn't found
+          if not resp.getheader('Content-Type', '').startswith('text/html'):
+            status = '###'
+            break
+          # assume characters in fragment are valid ASCII
+          rbody = resp.read().decode('ascii', 'ignore')
+          pairs = product(frags, (rbody,))
+          statuses = tuple(map(self._check_url_frag, pairs))
+          break
         elif 300 <= resp.status < 400:
           if redirs >= MAX_REDIRS:
             status = 'RRR'
@@ -267,30 +309,27 @@ class Checker():
 
     if start_url == url and not redirs:
       url = None
-    if frags:
-      if not statuses:
-        statuses = [status]*len(frags)
-      return statuses, url
-    return status, url
+    if not statuses:
+      statuses = (status,)*len(frags)
+    return statuses, url
 
-  def check_worker(self, q, r, running):
+  def check_worker(self, q, r, running, data):
 
+    local_html = data.get('local_html', '')
     try:
       while not q.empty() or running.value:
         try:
           url, frags = q.get(timeout=0.01)
-          if frags:
-            statuses, rurl = self.check_url(url, frags)
-            for frag, status in zip(frags, statuses):
-              data = (url + '#' + frag, (status, rurl))
-              while data:
-                try:
-                  r.put(data, False, 0.01)
-                  data = None
-                except Full:
-                  pass
-            continue
-          r.put((url, self.check_url(url)))
+          statuses, rurl = self.check_url(url, frags, local_html=local_html)
+          for frag, status in zip(frags, statuses):
+            data = (url + '#' + frag if frag else url, (status, rurl))
+            while data:
+              try:
+                r.put(data, False, 0.01)
+                data = None
+              except Full:
+                pass
+          continue
         except Empty:
           pass
     except KeyboardInterrupt:
@@ -315,22 +354,25 @@ class Checker():
     >>> c = Checker()
     >>> urls = ['http://example.com']
     >>> list(c._check_groupby(urls))
-    [('http://example.com', ())]
+    [('http://example.com', ('',))]
     >>> urls = [
     ...   'http://example.com',
     ...   'http://example.com/foo#bar1',
     ...   'http://example.com/foo#bar2',
+    ...   'http://example.com/foobar',
+    ...   'http://example.com/foobar#blah',
     ... ]
     >>> list(c._check_groupby(urls)) # doctest: +NORMALIZE_WHITESPACE
-    [('http://example.com', ()),
-     ('http://example.com/foo', ('bar1', 'bar2'))]
+    [('http://example.com', ('',)),
+     ('http://example.com/foo', ('bar1', 'bar2')),
+     ('http://example.com/foobar', ('', 'blah'))]
     >>> urls = ['#foo1', '#foo2']
     >>> list(c._check_groupby(urls))
     [('', ('foo1', 'foo2'))]
     """
     key = lambda item: item[0]
     for url, g in groupby(sorted(map(urldefrag, urls), key=key), key):
-      yield url, tuple(filter(None, (item[1] for item in g)))
+      yield url, tuple(item[1] for item in g)
 
   def check(self, f=None):
     """Check links
@@ -343,10 +385,12 @@ class Checker():
     r = Queue(self.QUEUE_SIZE)
     running = Value('b', 1)
     workers = []
+    manager = Manager()
+    datadict = manager.dict(self.data)
     for i in range(self.MAX_WORKERS):
       worker = Process(name='checker #%d' % i,
                        target=self.check_worker,
-                       args=(q, r, running))
+                       args=(q, r, running, datadict))
       worker.start()
       workers.append(worker)
 
@@ -391,7 +435,9 @@ class Checker():
     if self.json_filename:
       self.save_json(self.json_filename)
 
-  # =====
+  ###########################
+  # report output functions #
+  ###########################
 
   def color_status(self, status):
 
@@ -399,7 +445,7 @@ class Checker():
       return '\033[1;36m[%s]\033[0m' % status
     if status == '200':
       return '\033[1;32m[%s]\033[0m' % status
-    if status == 'SKP':
+    if status in ('SKP', 'SCH'):
       return '\033[1;33m[%s]\033[0m' % status
     else:
       return '\033[1;31m[%s]\033[0m' % status
@@ -410,15 +456,36 @@ class Checker():
     redir = link['redirection']
     print('%s %s' % (self.color_status(status), url), end='')
     if redir:
-      print(' \033[1;33m->\033[0m %s' % redir, end='')
+      print(' \033[1;33m\n   ->\033[0m %s' % redir, end='')
     print()
 
+  def num_len(self, n):
+    """Return length of n in string including thousands separator
+
+    >>> c = Checker()
+    >>> c.num_len(123)
+    3
+    >>> c.num_len(1234)
+    5
+    """
+    return len('{:,d}'.format(n))
+
+  def print_heading(self, text):
+
+    bar = '{0:{f}^{l}s}'.format('', f='=', l=len(text) + 4)
+    print('{0}\n= {1} =\n{0}\n'.format(bar, text))
+
+  def print_all(self):
+
+    self.print_report()
+    self.print_summary()
+
+  # report
+  #########
+
   def print_report(self):
 
-    print('==========')
-    print('* report *')
-    print('==========')
-    print()
+    self.print_heading('report')
 
     links = self.links
 
@@ -437,8 +504,7 @@ class Checker():
 
   def print_report_link(self, url, link):
 
-    status = link['status']
-    if status in (None, '200', 'SKP'):
+    if link['status'] in (None, '200', 'SCH', 'SKP'):
       return
     self.format_status(url, link)
     self.print_report_link_data(url, link)
@@ -447,22 +513,34 @@ class Checker():
 
     pass
 
-  # -----
+  # summary
+  ##########
 
   def print_summary(self):
 
-    print('===========')
-    print('* summary *')
-    print('===========')
-    print()
+    self.print_heading('summary')
 
     key = lambda link: link['status'] or '---'
+    data = []
     for status, g in groupby(sorted(self.links.values(), key=key), key=key):
-      links = list(g)
-      self.print_summary_status(status, links)
+      data.append((status, list(g)))
+    self.print_summary_status(data)
+    self.print_summary_footer()
+
+  def print_summary_status(self, data):
+
+    data2 = []
+    for status, links in data:
+      nlinks = sum(1 for l in links)
+      data2.append((status, nlinks))
+
+    l = max(self.num_len(item[1]) for item in data2)
+
+    for status, nlinks in data2:
+      print('{} {:{l},d} links'.format(self.color_status(status), nlinks, l=l))
+
+  def print_summary_footer(self):
+
+    print()
+    print('TOTAL {:,} links'.format(len(self.links)))
     print()
-
-  def print_summary_status(self, status, links):
-
-    nlinks = len(links)
-    print('%s %5d links' % (self.color_status(status), nlinks))
